/*
								+---------------------------------+
								|                                 |
								|  ***   Lightmap computer   ***  |
								|                                 |
								|  Copyright   -tHE SWINe- 2003  |
								|                                 |
								|            Light.cpp            |
								|                                 |
								+---------------------------------+
*/

#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "Matrix.h"
#include "Load_3ds.h"
#include "Light.h"

#define _Len(A) ((float)sqrt(A.x * A.x + A.y * A.y + A.z * A.z))
// spot dlku vektoru

#define _Normalize(A) {float s = _Len(A); A.x /= s; A.y /= s; A.z /= s; }
// vytvo jednotkov vektor

#define _Dot(A,B) (B.x * A.x + A.y * B.y + A.z * B.z)
// skalrn souin

#define _Cross(A,B,C) C.x = B.y * A.z - A.y * B.z; \
                     C.y = B.z * A.x - A.z * B.x; \
                     C.z = B.x * A.y - A.x * B.y;
// vektorov souin

/*
 *		=== Group creation ===
 */

int b_SameVertex(Vertex *p_vertex_a, Vertex *p_vertex_b)
{
    float f, d;

    f = (p_vertex_a->x - p_vertex_b->x);
    f *= f;
    d = (p_vertex_a->y - p_vertex_b->y);
    f += d * d;
    d = (p_vertex_a->z - p_vertex_b->z);
    // vzdlenost mezi vrcholy
    
    return ((f + d * d) < 0.01f);
}

int n_ClassifyVertex(Plane t_plane, Vertex *p_vertex)
{
	float f;

	if((f = p_vertex->x * t_plane.a + p_vertex->y * t_plane.b +
	   p_vertex->z * t_plane.c + t_plane.d) > .1f)
		return 1;
	else if(f < -.1f)
		return -1;
	return 0;
}

int b_Neighbour(Face *p_face_a, Face *p_face_b)
{
    int n_edge_a, n_edge_b;
    int n_edge_c, n_edge_d;
	Plane t_plane;

	t_plane = p_face_a->normal;
	t_plane.d = -(p_face_a->normal.a * p_face_b->vertex[0]->x +
		  p_face_a->normal.b * p_face_b->vertex[0]->y +
		  p_face_a->normal.c * p_face_b->vertex[0]->z);

    if(n_ClassifyVertex(t_plane, p_face_b->vertex[1]) ||
	   n_ClassifyVertex(t_plane, p_face_b->vertex[2]))
        return false;
    // facy nele v rovin

	t_plane = p_face_b->normal;
	t_plane.d = -(p_face_b->normal.a * p_face_a->vertex[0]->x +
		  p_face_b->normal.b * p_face_a->vertex[0]->y +
		  p_face_b->normal.c * p_face_a->vertex[0]->z);

    if(n_ClassifyVertex(t_plane, p_face_a->vertex[1]) ||
	   n_ClassifyVertex(t_plane, p_face_a->vertex[2]))
        return false;
    // facy nele v rovin (znova, vylepen konzistence)

    for(n_edge_a = 0; n_edge_a < 3; n_edge_a ++) {
        n_edge_b = (1 + n_edge_a) % 3;
        // indexy vertex hrany (a, b)

        for(n_edge_c = 0; n_edge_c < 3; n_edge_c ++) {
            n_edge_d = (n_edge_c + 1) % 3;
            // indexy vertex druh hrany (c, d)

            if((b_SameVertex(p_face_a->vertex[n_edge_a],
                            p_face_b->vertex[n_edge_c]) &&
               b_SameVertex(p_face_a->vertex[n_edge_b],
                            p_face_b->vertex[n_edge_d])) ||
               (b_SameVertex(p_face_a->vertex[n_edge_a],
                            p_face_b->vertex[n_edge_d]) &&
               b_SameVertex(p_face_a->vertex[n_edge_b],
                            p_face_b->vertex[n_edge_c])))
                return true;
        }
    }
    // zjiuje sousedn facy

    return false;
}

int Add_FaceToGroup(Group *p_group, Face *p_face)
{
    if(p_group->n_face_used == p_group->n_face_num) {
        if(p_group->n_face_num) {
            if((p_group->p_face = (Face**)realloc
               (p_group->p_face, (p_group->n_face_num + 4) * sizeof(Face*))) == NULL)
                return false;
            p_group->n_face_num += 4;
        } else {
            if((p_group->p_face = (Face**)malloc(2 * sizeof(Face*))) == NULL)
                return false;
            p_group->n_face_num = 2;
        }
    }
    // pokud je poteba, zvt zsobnk

    p_group->p_face[p_group->n_face_used ++] = p_face;
    // pid face

    return true;
}

int Check_GroupSize(Group *p_group, Face *p_face, float f_max_group_size)
{
	Vector v_new_min, v_new_max, v_new_size;
	int i;

	v_new_min = p_group->v_min;
	v_new_max = p_group->v_max;
	//
	for(i = 0; i < 3; i ++) {
		if(v_new_min.x > p_face->vertex[i]->x)
			v_new_min.x = p_face->vertex[i]->x;
		if(v_new_min.y > p_face->vertex[i]->y)
			v_new_min.y = p_face->vertex[i]->y;
		if(v_new_min.z > p_face->vertex[i]->z)
			v_new_min.z = p_face->vertex[i]->z;
		//
		if(v_new_max.x < p_face->vertex[i]->x)
			v_new_max.x = p_face->vertex[i]->x;
		if(v_new_max.y < p_face->vertex[i]->y)
			v_new_max.y = p_face->vertex[i]->y;
		if(v_new_max.z < p_face->vertex[i]->z)
			v_new_max.z = p_face->vertex[i]->z;
	}
	//
	v_new_size = v_new_max;
	v_new_size.x -= v_new_min.x;
	v_new_size.y -= v_new_min.y;
	v_new_size.z -= v_new_min.z;
	if(_Dot(v_new_size, p_group->v_u) > f_max_group_size ||
	   _Dot(v_new_size, p_group->v_v) > f_max_group_size) {
		if(v_new_size.x - (p_group->v_max.x - p_group->v_min.x) > .1f ||
		   v_new_size.y - (p_group->v_max.y - p_group->v_min.y) > .1f ||
		   v_new_size.z - (p_group->v_max.z - p_group->v_min.z) > .1f)
			return false;
		// pokud u byl moc velk, pidnm facu se nemusel zvtit
	}
	// pokud by po pidn facu byl group u moc velk ...

	p_group->v_max = v_new_max;
	p_group->v_min = v_new_min;
	// zupdatuje rozmry ...

	return true;
	// true = face se me pidat
}

void Init_Group(Group *p_group)
{
	Vector v_new_min, v_new_max;
	float v_f[3];
	int i;

	v_f[0] = (float)fabs(p_group->p_face[0]->normal.a);
	v_f[1] = (float)fabs(p_group->p_face[0]->normal.b);
	v_f[2] = (float)fabs(p_group->p_face[0]->normal.c);
	//
	for(i = 0; i < 2; i ++) {
		if(v_f[i] > v_f[(i + 1) % 3] &&
		   v_f[i] > v_f[(i + 2) % 3])
			break;
	}
	//
	p_group->n_plane = i;
	// najde nejvt sloku normly (rovinu na kterou se bude mapovat)

	memset(&p_group->v_u, 0, sizeof(Vector));
	memset(&p_group->v_v, 0, sizeof(Vector));
	switch(p_group->n_plane) {
	case Plane_yz: // mapovn po rovin yz
		p_group->v_u.z = 1;
		p_group->v_v.y = 1;
		break;
	case Plane_xz: // mapovn po rovin xz
		p_group->v_u.x = 1;
		p_group->v_v.z = 1;
		break;
	default: // mapovn po rovin xy
		p_group->v_u.x = 1;
		p_group->v_v.y = 1;
		break;
	}
	// najde u a v vektory mapovac roviny

	v_new_min.x = p_group->p_face[0]->vertex[0]->x;
	v_new_min.y = p_group->p_face[0]->vertex[0]->y;
	v_new_min.z = p_group->p_face[0]->vertex[0]->z;
	v_new_max.x = p_group->p_face[0]->vertex[0]->x;
	v_new_max.y = p_group->p_face[0]->vertex[0]->y;
	v_new_max.z = p_group->p_face[0]->vertex[0]->z;
	for(i = 1; i < 3; i ++) {
		if(v_new_min.x > p_group->p_face[0]->vertex[i]->x)
			v_new_min.x = p_group->p_face[0]->vertex[i]->x;
		if(v_new_min.y > p_group->p_face[0]->vertex[i]->y)
			v_new_min.y = p_group->p_face[0]->vertex[i]->y;
		if(v_new_min.z > p_group->p_face[0]->vertex[i]->z)
			v_new_min.z = p_group->p_face[0]->vertex[i]->z;
		//
		if(v_new_max.x < p_group->p_face[0]->vertex[i]->x)
			v_new_max.x = p_group->p_face[0]->vertex[i]->x;
		if(v_new_max.y < p_group->p_face[0]->vertex[i]->y)
			v_new_max.y = p_group->p_face[0]->vertex[i]->y;
		if(v_new_max.z < p_group->p_face[0]->vertex[i]->z)
			v_new_max.z = p_group->p_face[0]->vertex[i]->z;
	}
	//
	p_group->v_min = v_new_min;
	p_group->v_max = v_new_max;
	// najde min. a max. vektor fac z groupu ...
}

Group_Data *Group_Faces(World *p_world, float f_max_size)
{
    Group_Data *p_data;
    int i, j, k, l, m;
    int b_found;
	int b_skip;

    if((p_data = (Group_Data*)malloc(sizeof(Group_Data))) == NULL)
        return NULL;
    p_data->n_group_num = 0;
    p_data->p_group = NULL;
    // naalokuje strukturu
    
    for(i = 0; i < p_world->n_object_num; i ++) {
        for(j = 0; j < p_world->object[i].n_face_num; j ++) {
            for(k = 0, b_found = false; k < p_data->n_group_num; k ++) {
				if(!Check_GroupSize(&p_data->p_group[k],
				   &p_world->object[i].face[j], f_max_size))
					continue;

                for(l = 0; l < p_data->p_group[k].n_face_used; l ++) {
                    if(b_Neighbour(&p_world->object[i].face[j], p_data->p_group[k].p_face[l])) {
                        b_found = true;
                        if(!Add_FaceToGroup(&p_data->p_group[k],
                           &p_world->object[i].face[j]))
                            return false;
                        p_world->object[i].face[j].n_group = k;
                        // pid se do groupu ...

						for(k ++; k < p_data->n_group_num; k ++) {
							for(l = 0; l < p_data->p_group[k].n_face_used; l ++) {
								if(b_Neighbour(&p_world->object[i].face[j], p_data->p_group[k].p_face[l])) {
									// groupy p_world->object[i].face[j].n_group a k pijdou propojit ...

									b_skip = false;
									for(m = 0; m < p_data->p_group[k].n_face_used; m ++) {
										if(!Check_GroupSize(&p_data->p_group[p_world->object[i].face[j].n_group],
										   p_data->p_group[k].p_face[m], f_max_size)) {
											b_skip = true;
											continue;
										}
										if(!Add_FaceToGroup(&p_data->p_group[p_world->object[i].face[j].n_group],
										   p_data->p_group[k].p_face[m]))
											return false;
									}
									// pid vechny facy z groupu k ...

									if(!b_skip) {
										free((void*)p_data->p_group[k].p_face);
										p_data->p_group[k --] = p_data->p_group[-- p_data->n_group_num];
									}
									// smae group, pokud v nm nic nezstalo

									break;
								}
							}
						}
						// hled, jestli pat jet do njakho groupu ...

                        break;
                    }
                }
            }
			// hled jestli face pat do jednoho (vce) group ...

            if(!b_found) {
                if(!p_data->p_group) {
                    p_data->n_group_num = 1;
                    if(!(p_data->p_group = (Group*)malloc(sizeof(Group))))
                        return false;
                } else {
                    if(!(p_data->p_group = (Group*)realloc(p_data->p_group,
                       sizeof(Group) * ++ p_data->n_group_num)))
                        return false;
                }
                // realokuje groupy ...

				p_data->p_group[p_data->n_group_num - 1].n_id = p_data->n_group_num - 1;
                p_data->p_group[p_data->n_group_num - 1].n_face_num = 0;
                p_data->p_group[p_data->n_group_num - 1].n_face_used = 0;
                p_data->p_group[p_data->n_group_num - 1].p_face = 0;
				//
                if(!Add_FaceToGroup(&p_data->p_group[p_data->n_group_num - 1],
                   &p_world->object[i].face[j]))
                    return false;
                p_world->object[i].face[j].n_group = p_data->n_group_num - 1;
				// pid prvn face ...

				Init_Group(&p_data->p_group[p_data->n_group_num - 1]);
				// spot vektory
            }
        }
    }
    // hled sousedc facy

    return p_data;
}


/*
 *		=== Mapping coordinates ===
 */

Mapping_Info *Generate_Map_Coords(Group_Data *p_group_data)
{
	Mapping_Info *p_map_info;
	Vector v_min, v_max;
	Vector v_u, v_v;
	Group *p_group;
	int i, j, k;

	if(!(p_map_info = (Mapping_Info*)malloc(p_group_data->n_group_num *
	   sizeof(Mapping_Info))))
		return 0;
	// poet prvk vracenho pole je stejn jako poet group

	for(i = 0; i < p_group_data->n_group_num; i ++) {
		p_group = &p_group_data->p_group[i];

		v_min.x = p_group->p_face[0]->vertex[0]->x;
		v_min.y = p_group->p_face[0]->vertex[0]->y;
		v_min.z = p_group->p_face[0]->vertex[0]->z;
		v_max = v_min;
		//
		for(j = 0; j < p_group->n_face_used; j ++) {
			for(k = 0; k < 3; k ++) {
				if(v_min.x > p_group->p_face[j]->vertex[k]->x)
					v_min.x = p_group->p_face[j]->vertex[k]->x;
				if(v_min.y > p_group->p_face[j]->vertex[k]->y)
					v_min.y = p_group->p_face[j]->vertex[k]->y;
				if(v_min.z > p_group->p_face[j]->vertex[k]->z)
					v_min.z = p_group->p_face[j]->vertex[k]->z;
				//
				if(v_max.x < p_group->p_face[j]->vertex[k]->x)
					v_max.x = p_group->p_face[j]->vertex[k]->x;
				if(v_max.y < p_group->p_face[j]->vertex[k]->y)
					v_max.y = p_group->p_face[j]->vertex[k]->y;
				if(v_max.z < p_group->p_face[j]->vertex[k]->z)
					v_max.z = p_group->p_face[j]->vertex[k]->z;
			}
		}
		// najde min a max souadnice

		p_map_info[i].v_org = v_min;
		// poten vrchol (pomh zjistit, kde le jak bod na lightmap)

		v_u = p_group->v_u;
		v_v = p_group->v_v;
		// u a v vektory mapovac roviny

		p_map_info[i].n_plane = p_group->n_plane;
		// rovina, na kter se mapuje ,,,

		p_map_info[i].f_width = _Dot(v_max, v_u) - _Dot(v_min, v_u);
		p_map_info[i].f_height = _Dot(v_max, v_v) - _Dot(v_min, v_v);
		// me se pout normln mnus, protoe vdycky
		// bude aktivn jen jedna souadnice. normln by
		// se pouila vzdlenost dvou vektor. jinak pot
		// ku a vku (u 2D) mapovacho obdlnku

		p_map_info[i].v_u = v_u;
		p_map_info[i].v_v = v_v;
		// vektory hran

		for(j = 0; j < p_group->n_face_used; j ++) {
			for(k = 0; k < 3; k ++) {
				p_group->p_face[j]->vertex[k]->lu =
					((p_group->p_face[j]->vertex[k]->x - v_min.x) * v_u.x +
					(p_group->p_face[j]->vertex[k]->y - v_min.y) * v_u.y +
					(p_group->p_face[j]->vertex[k]->z - v_min.z) * v_u.z) /
					p_map_info[i].f_width;
				p_group->p_face[j]->vertex[k]->lv =
					((p_group->p_face[j]->vertex[k]->x - v_min.x) * v_v.x +
					(p_group->p_face[j]->vertex[k]->y - v_min.y) * v_v.y +
					(p_group->p_face[j]->vertex[k]->z - v_min.z) * v_v.z) /
					p_map_info[i].f_height;
			}
		}
		// spot koordinty lightmapy ve vech vertexech fac, patcch do groupu
	}

	return p_map_info;
}

/*
 *		=== Lightmap tile packing ===
 */

TLightmapTex *p_Lightmap(int n_width, int n_height)
{
	TLightmapTex *p_lightmap;

	if(!(p_lightmap = (TLightmapTex*)malloc(sizeof(TLightmapTex))))
		return 0;
	// naalokuje data pro lightmapu ...

	p_lightmap->n_width = n_width;
	p_lightmap->n_height = n_height;
	// nastav rozmry lightmapy ...

	if(!(p_lightmap->p_frontline = (int*)malloc(n_width * sizeof(int))))
		return 0;
	memset(p_lightmap->p_frontline, 0, n_width * sizeof(int));
	// naalokuje a vynuluje data pro vkldn dlk ...

	if(!(p_lightmap->p_bytes = (unsigned __int32*)malloc(n_width *
	   n_height * sizeof(unsigned __int32))))
		return 0;
	memset(p_lightmap->p_bytes, 0, n_width * n_height * sizeof(unsigned __int32));
	// naalokuje buffer pro texely a vymae ho ...

	return p_lightmap;
}

int b_InsertTile(int n_width, int n_height, TLightmapTex *p_lightmap,
				 int *p_x, int *p_y, Matrix *p_coord_transform)
{
	int i, j, n_max;

	n_width += 2;
	n_height += 2;
	// pidme kad lightmap jednopixelov rmeek ...

	if(n_width > p_lightmap->n_width || n_height > p_lightmap->n_height)
		return false;
	// nevejde se ...

	for(i = 0, j = 0, n_max = 0; i < p_lightmap->n_width; i ++) {
		if(p_lightmap->p_frontline[i] + n_height >= p_lightmap->n_height) {
			n_max = 0;
			j = 0;
			continue;
		} else {
			if(n_max < p_lightmap->p_frontline[i])
				n_max = p_lightmap->p_frontline[i];
			j ++;
		}
		// hled, jestli dlek nebude pesahovat z lightapy,
		// zrove hled nejvy uloen dlek v oblasti ...

		if(j == n_width) {
			for(*p_x = (i -= j - 1); i < (*p_x) + j; i ++)
				p_lightmap->p_frontline[i] = n_max + n_height;
			// vypln data kde bude dal dlek ...

			*p_y = n_max;
			// souadnice kde bude dlek uloen ...

			(*p_x) ++;
			(*p_y) ++;
			// posuneme o jeden pixel doprava a dol,
			// aby dlek byl uprosted svho rmeku

			n_width -= 2;
			n_height -= 2;

			Matrix_Init(p_coord_transform);
			//
			(*p_coord_transform)[0][0] *= (float)n_width / (float)p_lightmap->n_width;
			(*p_coord_transform)[0][1] *= (float)n_width / (float)p_lightmap->n_width;
			(*p_coord_transform)[0][2] *= (float)n_width / (float)p_lightmap->n_width;
			//
			(*p_coord_transform)[1][0] *= (float)n_height / (float)p_lightmap->n_height;
			(*p_coord_transform)[1][1] *= (float)n_height / (float)p_lightmap->n_height;
			(*p_coord_transform)[1][2] *= (float)n_height / (float)p_lightmap->n_height;
			//
			(*p_coord_transform)[3][0] = (float)(*p_x - .5f) / (float)p_lightmap->n_width;
			(*p_coord_transform)[3][1] = (float)(*p_y - .5f) / (float)p_lightmap->n_height;
			// vytvo matici, kter se pouije k petransformovn souadnic do nov lightmapy

			return true;
		}
	}
	// hled kam by se mohla lightmapa vejt ...

	return false;
}

/*
 *		=== Lightmap tracing ===
 */

// vrt true, pokud paprsek narazil na face
int Cast_Ray(Vector v_vector_a, Vector v_vector_b, Face *p_face)
{
	Vector mid, v1, v2, pl;
    float a, b, t, d;

    a = v_vector_a.x * p_face->normal.a + v_vector_a.y * p_face->normal.b +
        v_vector_a.z * p_face->normal.c + p_face->normal.d;
    b = v_vector_b.x * p_face->normal.a + v_vector_b.y * p_face->normal.b +
        v_vector_b.z * p_face->normal.c + p_face->normal.d;
    // vzdlenosti od roviny

    if(a * b > 0)
        return false;
    // pokud jsou oba body na stejn stran

	if(fabs(a) < .1f || fabs(b) < .1f)
		return false;
	// pokud le onplane (jeden vdycky bude, protoe le
	// na polygonu, kam je lightmapa mapovan)

    t = -a / (b - a);
    mid.x = v_vector_a.x + t * (v_vector_b.x - v_vector_a.x);
    mid.y = v_vector_a.y + t * (v_vector_b.y - v_vector_a.y);
    mid.z = v_vector_a.z + t * (v_vector_b.z - v_vector_a.z);
    // spot prsek

    v1.x = -p_face->normal.a;
    v1.y = -p_face->normal.b;
    v1.z = -p_face->normal.c;
    // prvn vektor (pro vechny roviny stejn)

    v2.x = p_face->vertex[1]->x - p_face->vertex[2]->x;
    v2.y = p_face->vertex[1]->y - p_face->vertex[2]->y;
    v2.z = p_face->vertex[1]->z - p_face->vertex[2]->z;
    // prvn hranov vektor

    _Cross(v1, v2, pl);
    _Normalize(pl);
    d = - (pl.x * p_face->vertex[2]->x +
           pl.y * p_face->vertex[2]->y +
           pl.z * p_face->vertex[2]->z);
    // spot rovinu

    if(mid.x * pl.x + mid.y * pl.y + mid.z * pl.z + d < -1e-3)
        return false;
    // otestuje prsek

    v2.x = p_face->vertex[0]->x - p_face->vertex[1]->x;
    v2.y = p_face->vertex[0]->y - p_face->vertex[1]->y;
    v2.z = p_face->vertex[0]->z - p_face->vertex[1]->z;
    // druh vektor

    _Cross(v1, v2, pl);
    _Normalize(pl);
    d = - (pl.x * p_face->vertex[1]->x +
           pl.y * p_face->vertex[1]->y +
           pl.z * p_face->vertex[1]->z);
    // spot rovinu

    if(mid.x * pl.x + mid.y * pl.y + mid.z * pl.z + d < -1e-3)
        return false;
    // otestuje prsek

    v2.x = p_face->vertex[2]->x - p_face->vertex[0]->x;
    v2.y = p_face->vertex[2]->y - p_face->vertex[0]->y;
    v2.z = p_face->vertex[2]->z - p_face->vertex[0]->z;
    // treti vektor

    _Cross(v1, v2, pl);
    _Normalize(pl);
    d = - (pl.x * p_face->vertex[0]->x +
           pl.y * p_face->vertex[0]->y +
           pl.z * p_face->vertex[0]->z);
    // spot posledn rovinu

    if(mid.x * pl.x + mid.y * pl.y + mid.z * pl.z + d < -1e-3)
        return false;
    // otestuje prsek

    return true;
}

// vrt true, pokud paprsek projde od zatku k cli
int Cast_Ray(Vector v_vector_a, Vector v_vector_b, World *p_world)
{
	int i, j;

	for(i = 0; i < p_world->n_object_num; i ++) {
		for(j = 0; j < p_world->object[i].n_face_num; j ++) {
			if(p_world->object[i].face[j].normal.a * v_vector_a.x +
			   p_world->object[i].face[j].normal.b * v_vector_a.y +
			   p_world->object[i].face[j].normal.c * v_vector_a.z +
			   p_world->object[i].face[j].normal.d > 0)
				continue;

			if(Cast_Ray(v_vector_a, v_vector_b, &p_world->object[i].face[j]))
				return false;
		}
	}

	return true;
}

void Raytrace_Tile(Group *p_group, Mapping_Info *p_mapping_info, World *p_world,
				   int n_x, int n_y, int n_h, int n_w, int n_p_w,
				   unsigned __int32 *p_buffer, unsigned __int32 n_ambient)
{
	Vector v_texel_pos;
	Vector v_pos_v;
	float r, g, b;
	Vector v_ray;
	int x, y;
	float a;
	int j;

	for(y = 0; y < n_h; y ++) {
		v_pos_v = p_mapping_info->v_org;
		v_pos_v.x += p_mapping_info->v_v.x * (float)y /
			(float)n_h * p_mapping_info->f_height;
		v_pos_v.y += p_mapping_info->v_v.y * (float)y /
			(float)n_h * p_mapping_info->f_height;
		v_pos_v.z += p_mapping_info->v_v.z * (float)y /
			(float)n_h * p_mapping_info->f_height;
		//
		for(x = 0; x < n_w; x ++) {
			v_texel_pos.x = p_mapping_info->v_u.x * (float)x /
				(float)n_w * p_mapping_info->f_width + v_pos_v.x;
			v_texel_pos.y = p_mapping_info->v_u.y * (float)x /
				(float)n_w * p_mapping_info->f_width + v_pos_v.y;
			v_texel_pos.z = p_mapping_info->v_u.z * (float)x /
				(float)n_w * p_mapping_info->f_width + v_pos_v.z;
			// spot pozici texelu lightmapy ve worldspace

			switch(p_mapping_info->n_plane) {
			case Plane_xy:
				v_texel_pos.z = -(v_texel_pos.x * p_group->p_face[0]->normal.a +
								  v_texel_pos.y * p_group->p_face[0]->normal.b +
								  p_group->p_face[0]->normal.d) / p_group->p_face[0]->normal.c;
				break;
			case Plane_xz:
				v_texel_pos.y = -(v_texel_pos.x * p_group->p_face[0]->normal.a +
								  v_texel_pos.z * p_group->p_face[0]->normal.c +
								  p_group->p_face[0]->normal.d) / p_group->p_face[0]->normal.b;
				break;
			case Plane_yz:
				v_texel_pos.x = -(v_texel_pos.y * p_group->p_face[0]->normal.b +
								  v_texel_pos.z * p_group->p_face[0]->normal.c +
								  p_group->p_face[0]->normal.d) / p_group->p_face[0]->normal.a;
				break;
			}
			// promtne pozici texelu z naeho obdlnku do roviny, ve kter le facy ...

			r = (float)((n_ambient >> 16) & 0xff) / 256;
			g = (float)((n_ambient >> 8) & 0xff) / 256;
			b = (float)(n_ambient & 0xff) / 256;
			// tady by nemly bt nuly, ale ambient ...

			for(j = 0; j < p_world->n_light_num; j ++) {
				v_ray.x = p_world->light[j].v_position.x - v_texel_pos.x;
				v_ray.y = p_world->light[j].v_position.y - v_texel_pos.y;
				v_ray.z = p_world->light[j].v_position.z - v_texel_pos.z;
				//
				a = - p_group->p_face[0]->normal.a * v_ray.x +
					- p_group->p_face[0]->normal.b * v_ray.y +
					- p_group->p_face[0]->normal.c * v_ray.z;
				a /= _Len(v_ray);
				// spot hel, kter svr vektor texel - svtlo s normlou

				if(a <= 0)
					continue;

				/*if(p_world->light[j].f_outer_range > p_world->light[j].f_inner_range &&
				   _Len(v_ray) > p_world->light[j].f_inner_range) {
					a *=  1 - (_Len(v_ray) - p_world->light[j].f_inner_range) /
						(p_world->light[j].f_outer_range - p_world->light[j].f_inner_range);
				}*/
				// spot zeslaben svtla vzdlenost, pokud je nastaven

				if(a > 0 && (/*!p_world->light[j].b_cast_shadows ||*/
				   Cast_Ray(p_world->light[j].v_position, v_texel_pos, p_world))) {
					a = (a > 1)? p_world->light[j].f_multiplier :
						p_world->light[j].f_multiplier * a;
					// zapot nsobi svtla (monost 3ds)

					r += p_world->light[j].v_color.x * a;
					b += p_world->light[j].v_color.y * a;
					g += p_world->light[j].v_color.z * a;
					// spot svtlost
				}
				// pokud bude svtlo vidt, a bu nevrh stny, nebo na texelu nen stn
			}
			// pot svtlost ...

			r = (r > 1)? 255 : r * 255;
			g = (g > 1)? 255 : g * 255;
			b = (b > 1)? 255 : b * 255;
			// sjede rgb do <0, 1>

			p_buffer[(x + n_x) + (y + n_y) * n_p_w] =
				((int)r << 16) + ((int)g << 8) + (int)b;
			// pevede barvu do RGB a ulo do pslunho dlku v textue ...
		}
	}
}
