/*
                                +---------------------------------+
                                |                                 |
                                |       ***   Load3DS   ***       |
                                |                                 |
                                |  Copyright   -tHE SWINe- 2002  |
                                |                                 |
                                |            Load3DS.h            |
                                |                                 |
                                +---------------------------------+
*/

#ifndef __3DS_INCLUDED
#define __3DS_INCLUDED

struct Plane {
    float a, b, c, d;
};

struct Vector {
    float x, y, z;
};

struct Vertex {
	int n_id;
	//
    float x, y, z;
    float u, v;
	float lu, lv;
    //
    Vector normal;
};

typedef float Matrix[4][4];

struct Mip_Map {
	int n_Width, n_Height;
	unsigned char *texture;
	int A_1, A_2, A_3, B_1;
};

struct Texture {
	int n_Width;
	int n_Height;
	int n_mip_num_x;
	int n_mip_num_y;
	unsigned __int32 *palette;
	Mip_Map *mip_map;
};

struct Material {
    int n_id;
    char *name;
	//
	Texture *texture;
};

struct Light {
    int n_id;
    char *name;
    //
    int b_spot;
    int b_off;
    int b_cast_shadows;
    //
    Vector v_position;
    Vector v_target;
    //
    float f_hotspot;
    float f_faloff;
    //
    float f_inner_range;
    float f_outer_range;
    float f_multiplier;
    //
    Vector v_color;
    int r, g, b;
};

struct Face {
    int n_id, n_used;
    int n_material;
	int n_lightmap;
	int n_group;
	//
	Face *n_neighbour[3];
    //
    Vertex *vertex[3];
    //
    Plane normal;
};

struct Object {
    int n_id;
    char *name;
    //
    int n_vertex_num;
    int n_face_num;
    //
    Face *face;
    Vertex *vertex;
    //
    Matrix m_object;
    //
    int n_parent;
    int n_pos_num;
    int n_rot_num;
    //
    float *pt, *px, *py, *pz;
    float *qt, *qw, *qx, *qy, *qz;
};

struct Camera {
    int n_id;
    char *name;
    //
    Matrix m_camera;
    //
    float f_pos_x, f_pos_y, f_pos_z;
    float f_tar_x, f_tar_y, f_tar_z;
    float f_bank_angle;
    float f_focus; 
    //
    int n_target_num;
    int n_position_num;
    float *tt, *tx, *ty, *tz;
    float *pt, *px, *py, *pz;
};

struct World {
    int n_start_frame;
    int n_end_frame;
    //
    int n_object_num;
    int n_camera_num;
    int n_light_num;
    int n_material_num;
    //
    Object *object;
    Camera *camera;
    Light *light;
    Material *material;
};


World *Load_3DS(char *file_name);
Plane Make_Plane(Face *p_face);

#endif
