/*
								+---------------------------------+
								|                                 |
								|  ***   Lightmap computer   ***  |
								|                                 |
								|  Copyright   -tHE SWINe- 2003  |
								|                                 |
								|            Main.cpp             |
								|                                 |
								+---------------------------------+
*/

#include <windows.h>
#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "Matrix.h"
#include "Load_3ds.h"
#include "Light.h"
#include "Main.h"

#define Pi (3.141592653589794626433832795028841931f)

int	Save_TrueColor_BMP(char *filename, TBmp *bmp)
{
	unsigned long ByteCount;
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER bmih;
	unsigned char *lpbBuf;
	unsigned __int32 bBuf;
	unsigned int BuffLen;
	int i, j, x;
	FILE *BMPfp;

	BuffLen = ((3 * bmp->n_width + 3) >> 2) << 2;
	lpbBuf =  (unsigned char*) malloc(BuffLen);
	// compute lenght of line-buffer

	if((BMPfp = fopen(filename, "wb")) == NULL)
		return 0;
	// opens file

	bmfh.bfType = ('B'|('M'<<8)); // "BM"
	bmfh.bfSize = 0;              // necessary to modify
	bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;
	bmfh.bfOffBits = sizeof(BITMAPFILEHEADER) +
		sizeof(BITMAPINFOHEADER);
	bmih.biSize = sizeof(BITMAPINFOHEADER);
	bmih.biWidth = bmp->n_width;
	bmih.biHeight = bmp->n_height;
	bmih.biPlanes = 1;
	bmih.biBitCount = 24;	      // True Color
	bmih.biCompression = BI_RGB;
	bmih.biSizeImage = 0;         // necessary to modify
	bmih.biXPelsPerMeter = 96;
	bmih.biYPelsPerMeter = 96;
	bmih.biClrUsed = 0;
	bmih.biClrImportant = 0;
	// create header

	fwrite(&bmfh, sizeof(BITMAPFILEHEADER), 1, BMPfp);
	fwrite(&bmih, sizeof(BITMAPINFOHEADER), 1, BMPfp);
	// writing header

	ByteCount = 0;
	for(i = bmp->n_height - 1; i >= 0; i --) {
		for(j = 0, x = 0; j < bmp->n_width; j ++, x ++) {
			bBuf = bmp->p_bytes[i * bmp->n_width + j];
			lpbBuf[x * 3] = (char)((bBuf >> 16) & 0xff);
			lpbBuf[x * 3 + 1] = (char)((bBuf >> 8) & 0xff);
			lpbBuf[x * 3 + 2] = (char)(bBuf & 0xff);
			// saved in BGR, not RGB
			ByteCount += 3;
		}

		if(fwrite(lpbBuf, 1, BuffLen, BMPfp) != BuffLen) {
			free(lpbBuf);
			return 0;
		}
	}

	// repair header
	bmih.biSizeImage = ByteCount;
	bmfh.bfSize = sizeof(BITMAPFILEHEADER) +
	   ByteCount + sizeof(BITMAPINFOHEADER);
	fseek(BMPfp, 0, SEEK_SET);
	fwrite(&bmfh, sizeof(BITMAPFILEHEADER), 1, BMPfp);
	fwrite(&bmih, sizeof(BITMAPINFOHEADER), 1, BMPfp);

	fclose(BMPfp);

	free(lpbBuf);

	return 1;
}

Object *p_IsolateVertices(Object *p_object)
{
	Object *p_new_object;
	int i, j;

	if(!(p_new_object = (Object*)malloc(sizeof(Object))))
		return 0;
	//
	p_new_object->n_face_num = p_object->n_face_num;
	p_new_object->n_vertex_num = p_object->n_face_num * 3;
	//
	if(!(p_new_object->face = (Face*)malloc(sizeof(Face) * p_object->n_face_num)))
		return 0;
	if(!(p_new_object->vertex = (Vertex*)malloc(sizeof(Vertex) * p_new_object->n_vertex_num)))
		return 0;
	// naalokuje objekt, vertexy a facy

	for(i = 0; i < p_object->n_vertex_num; i ++)
		p_object->vertex[i].n_id = i;
	// ulom si do kadho vertexu jeho id. podle
	// toho potom budu zase vertexy dvat dohromady ...

	for(i = 0; i < p_object->n_face_num; i ++) {
		p_new_object->face[i] = p_object->face[i];
		//
		for(j = 0; j < 3; j ++) {
			p_new_object->vertex[i * 3 + j] = *p_object->face[i].vertex[j];
			p_new_object->face[i].vertex[j] = &p_new_object->vertex[i * 3 + j];
		}
		// zkopruje face a pehod pointery na vertexy, jejich obsah taky zkopruje
	}

	return p_new_object;
}

Vertex *p_FindVertex(Vertex *p_vertex, Object *p_object)
{
	int i;

	for(i = 0; i < p_object->n_vertex_num; i ++) {
		if(p_object->vertex[i].n_id == p_vertex->n_id && p_vertex != &p_object->vertex[i] &&
		   p_object->vertex[i].lu == p_vertex->lu && p_object->vertex[i].lv == p_vertex->lv)
			return &p_object->vertex[i];
	}
	// hled vertex, rozdln pointerem od p_vertex, se stejnm
	// id (= stejn pozice) a stejnmi souadnicemi lightmapy

	return 0;
}

Object *p_MergeVertices(Object *p_object)
{
	Object *p_new_object;
	Vertex *p_vertex;
	int i, j;

	if(!(p_new_object = (Object*)malloc(sizeof(Object))))
		return 0;
	//
	p_new_object->vertex = 0;
	p_new_object->n_vertex_num = 0;
	// uvoln vertexy objektu ...

	p_new_object->n_face_num = p_object->n_face_num;
	//
	if(!(p_new_object->face = (Face*)malloc(sizeof(Face) * p_object->n_face_num)))
		return 0;
	// naalokuje facy ...

	for(i = 0; i < p_object->n_face_num; i ++) {
		p_new_object->face[i] = p_object->face[i];
		//
		for(j = 0; j < 3; j ++) {
			if(!(p_vertex = p_FindVertex(p_object->face[i].vertex[j], p_object))) {
				if(!(p_new_object->vertex = (Vertex*)realloc(p_new_object->vertex,
				   ++ p_new_object->n_vertex_num * sizeof(Vertex))))
					return 0;
				p_new_object->vertex[p_new_object->n_vertex_num - 1] = *p_object->face[i].vertex[j];
				p_new_object->face[i].vertex[j] = (Vertex*)(p_new_object->n_vertex_num - 1);
			} else
				p_new_object->face[i].vertex[j] = (Vertex*)(p_vertex - p_object->vertex);
		}
	}
	// vytvo vertex list a ulo msto pointer na vertexy
	// jejich indexy (jejich adresa se pi realloc mn)

	for(i = 0; i < p_object->n_face_num; i ++) {
		for(j = 0; j < 3; j ++) {
			p_new_object->face[i].vertex[j] = p_new_object->vertex +
				(int)p_new_object->face[i].vertex[j];
		}
	}
	// zamn indexy za pointery ...

	return p_new_object;
}

// vrt index polka ...
inline int n_Index(int cx, int cy, int x, int y, int w, int h, int pw)
{
	if(cx < x)
		cx = x;
	else if(cx >= x + w)
		cx = x + w - 1;
	if(cy < y)
		cy = y;
	else if(cy >= y + h)
		cy = y + h - 1;

	return cx + cy * pw;
}

int Blur_Window(int x, int y, int w, int h, int pw, int n_size, unsigned __int32 *p_buffer)
{
	float f_r, f_g, f_b, f_size, f_weight;
	unsigned __int32 *p_temp_buffer;
	unsigned __int32 pix;
	int cx, cy;
	int u, v;

	if(!(p_temp_buffer = (unsigned __int32*)malloc(w * h * sizeof(unsigned __int32))))
		return false;

	for(v = y; v < y + h; v ++) {
		for(u = x; u < x + w; u ++) {
			f_r = 0;
			f_g = 0;
			f_b = 0;
			f_size = 0;
			//
			for(cx = u - n_size; cx <= u + n_size; cx ++) {
				for(cy = v - n_size; cy <= v + n_size; cy ++) {
					pix = p_buffer[n_Index(cx, cy, x, y, w, h, pw)];
					f_weight = 1 / (1 + (float)sqrt((cx - u) * (cx - u) + (cy - v) * (cy - v)));
					//
					f_r += (pix & 0xff0000) * f_weight;
					f_g += (pix & 0x00ff00) * f_weight;
					f_b += (pix & 0x0000ff) * f_weight;

					f_size += f_weight;
				}
			}
			//
			f_size = 1 / f_size;

			f_r *= f_size;
			f_g *= f_size;
			f_b *= f_size;

			p_temp_buffer[u - x + (v - y) * w] = (((int)f_r) & 0xff0000) |
				(((int)f_g) & 0xff00) | (((int)f_b) & 0xff);
		}
	}
	// rozmae obrzek do doasnhu bufferu ...

	for(v = y; v < y + h; v ++) {
		for(u = x; u < x + w; u ++)
			p_buffer[u + v * pw] = p_temp_buffer[u - x + (v - y) * w];
	}
	// zkopruje data z doasnhu bufferu ...

	free((void*)p_temp_buffer);

	return true;
}

void main(int argc, char *argv[])
{
	TLightmapTex **p_lightmap_page;
	unsigned __int32 n_ambient;
	int n_lightmap_page_num;
	Mapping_Info *p_m_info;
	Matrix m_coord_trans;
	Group_Data *p_g_data;
	int n_cur_w, n_cur_h;
	int n_cur_x, n_cur_y;
	float f_tr_u, f_tr_v;
	int n_lightmap_size;
	int i, j, k, l, n;
	Object *p_object;
	float f_quality;
	World *p_svet;
	char str[256];
	TBmp tmp_bmp;
	int n_blur;
	FILE *fw;

	printf("                                 === Light ===\n\n");

	if(argc < 2) {
		printf("Must be run with parameter :\n\nLight.exe [file.3ds]"
			" <-qx> <-ay> <-sz> <-bu>\n\n"
			"    where x is (float) quality of lightmaps (0 to inf),\n"
			"    y is ambient color (6 hex digits, for example 2b2b2b),\n"
			"    z is resolution of lightmap pages (1 to inf, rather powers of two) and\n"
			"    u is blur strength (0 to inf, integer)\n");
		getchar();
		return;
	}

	printf("- Loading world -\n\n");
	if((p_svet = Load_3DS(argv[1])) == NULL) {
		printf("    - Failure\n");
		getchar();
		return;
	}

	n_lightmap_size = 512;
	f_quality = 1;
	n_ambient = 0;
	n_blur = 0;
	//
	if(argc > 2) {
		for(i = 2; i < argc; i ++) {
			if(strstr(argv[i], "-q") || strstr(argv[i], "-Q"))
				f_quality = (float)atof(argv[i] + 2);
			else if(strstr(argv[i], "-s") || strstr(argv[i], "-S"))
				n_lightmap_size = atoi(argv[i] + 2);
			else if(strstr(argv[i], "-a") || strstr(argv[i], "-A"))
				sscanf(argv[i] + 2, "%x", &n_ambient);
			else if(strstr(argv[i], "-b") || strstr(argv[i], "-B"))
				n_blur = atoi(argv[i] + 2);
		}
	}
	// pete parametry ...

	printf("_lightmap quality is %f\n", f_quality);
	printf("_lightmap ambient is #%06x\n", n_ambient);
	printf("_lightmap resolution is %d\n", n_lightmap_size);
	printf("_lightmap blur is %d\n\n", n_blur);
	// zparsuje kvalitu (poet paprsk na pixel)

	printf("- Creating lightmaps -\n    - Isolating vertices\n");
	for(i = 0; i < p_svet->n_object_num; i ++) {
		if(!(p_object = p_IsolateVertices(&p_svet->object[i]))) {
			printf("    - Failure\n");
			return;
			getchar();
		}
		// vytvo objekt s izolovanmi vertexy (kad face m svoje vlastn vertexy)

		free((void*)p_svet->object[i].face);
		free((void*)p_svet->object[i].vertex);
		// uvoln stvajc vertexy a facy

		p_svet->object[i].face = p_object->face;
		p_svet->object[i].n_vertex_num = p_object->n_vertex_num;
		p_svet->object[i].vertex = p_object->vertex;
		// nastav nov vertexy a jejich poet a facy (poet fac se nezmn)

		free((void*)p_object);
		// uvoln pomocn objekt
	}
	// izoluje vertexy ...

	printf("    - Creating groups\n");
	if((p_g_data = Group_Faces(p_svet, (float)n_lightmap_size * f_quality)) == NULL) {
		printf("    - Failure\n");
		getchar();
		return;
	}
	// vytvo groupy fac, kad group bude mt jednu lightmapu

	printf("    - Generating coordinates\n");
	if((p_m_info = Generate_Map_Coords(p_g_data)) == NULL) {
		printf("    - Failure\n");
		getchar();
		return;
	}
	// spot mapovac souadnice (u, v pro lightmapy)

	printf("    - Generating texture pages\n");

	n_lightmap_page_num = 1;
	if(!(p_lightmap_page = (TLightmapTex**)malloc(sizeof(TLightmapTex*)))) {
		printf("    - Failure\n");
		return;
		getchar();
	}
	// naalokuje pam pro list se strnkami ...
	
	if(!(p_lightmap_page[0] = p_Lightmap(n_lightmap_size, n_lightmap_size))) {
		printf("    - Failure\n");
		return;
		getchar();
	}
	// naalokuje prvn strnku ...

	for(i = 0; i < p_g_data->n_group_num; i ++) {
		n_cur_w = (int)ceil(p_m_info[i].f_width * f_quality) + 1;
		n_cur_h = (int)ceil(p_m_info[i].f_height * f_quality) + 1;
		//
		n_cur_w = (n_cur_w > n_lightmap_size)? n_lightmap_size : n_cur_w;
		n_cur_h = (n_cur_h > n_lightmap_size)? n_lightmap_size : n_cur_h;
		// ur velikosti mapy pro group i

		for(j = 0; j < n_lightmap_page_num; j ++) {
			if(b_InsertTile(n_cur_w, n_cur_h, p_lightmap_page[j],
			   &n_cur_x, &n_cur_y, &m_coord_trans)) {
				for(k = 0; k < p_g_data->p_group[i].n_face_used; k ++) {
					for(l = 0; l < 3; l ++) {
						f_tr_u = m_coord_trans[0][0] *
								 p_g_data->p_group[i].p_face[k]->vertex[l]->lu +
								 m_coord_trans[0][1] *
								 p_g_data->p_group[i].p_face[k]->vertex[l]->lv +
								 m_coord_trans[3][0];
						f_tr_v = m_coord_trans[1][0] *
								 p_g_data->p_group[i].p_face[k]->vertex[l]->lu +
								 m_coord_trans[1][1] *
								 p_g_data->p_group[i].p_face[k]->vertex[l]->lv +
								 m_coord_trans[3][1];
						//
						p_g_data->p_group[i].p_face[k]->vertex[l]->lu = f_tr_u;
						p_g_data->p_group[i].p_face[k]->vertex[l]->lv = f_tr_v;
						//
						p_g_data->p_group[i].p_face[k]->n_lightmap = j;
					}
				}
				// natransformuje souadnice z naeho pomyslnho obdlnku do texturov strnky

				Raytrace_Tile(&p_g_data->p_group[i], &p_m_info[i], p_svet,
					n_cur_x, n_cur_y, n_cur_h, n_cur_w, p_lightmap_page[j]->n_width,
					p_lightmap_page[j]->p_bytes, n_ambient);
				// vypot svtlost jednoho dlku pmo v lightmap ...

				putchar('.');

				for(k = n_cur_x; k < n_cur_x + n_cur_w; k ++) {
					p_lightmap_page[j]->p_bytes[k +
						p_lightmap_page[j]->n_width * (n_cur_y - 1)] =
						p_lightmap_page[j]->p_bytes[k +
						p_lightmap_page[j]->n_width * n_cur_y];
					p_lightmap_page[j]->p_bytes[k +
						p_lightmap_page[j]->n_width * (n_cur_y + n_cur_h)] =
						p_lightmap_page[j]->p_bytes[k +
						p_lightmap_page[j]->n_width * (n_cur_y + n_cur_h - 1)];
				}
				// zkopruje rmeek horizontln ...

				for(l = n_cur_y - 1; l < n_cur_y + n_cur_h + 1; l ++) {
					p_lightmap_page[j]->p_bytes[n_cur_x - 1 +
						p_lightmap_page[j]->n_width * l] =
						p_lightmap_page[j]->p_bytes[n_cur_x +
						p_lightmap_page[j]->n_width * l];
					p_lightmap_page[j]->p_bytes[n_cur_x + n_cur_w +
						p_lightmap_page[j]->n_width * l] =
						p_lightmap_page[j]->p_bytes[n_cur_x - 1 + n_cur_w +
						p_lightmap_page[j]->n_width * l];
				}
				// zkopruje rmeek vertikln

				Blur_Window(n_cur_x, n_cur_y, n_cur_w, n_cur_h,
					p_lightmap_page[j]->n_width, n_blur, p_lightmap_page[j]->p_bytes);
				// rozmzne lightmapu
			} else if(j == n_lightmap_page_num - 1) {
				if(!(p_lightmap_page = (TLightmapTex**)realloc(p_lightmap_page,
				   ++ n_lightmap_page_num * sizeof(TLightmapTex*)))) {
					printf("    - Failure\n");
					return;
					getchar();
				}
				//
				if(!(p_lightmap_page[n_lightmap_page_num - 1] =
				   p_Lightmap(n_lightmap_size, n_lightmap_size))) {
					printf("    - Failure\n");
					return;
					getchar();
				}
				// naalokuje dal strnku, pokud se textura nevejde do dn z pedchozch 
			}
		}
		// pid lightmapu do texturov strnky
	}
	// postupn pid vechny groupy do textur, spot jejich
	// lightmapy a udl okolo nich jednopixelovej rmeek

	printf("\n    - Merging vertices\n");
	/*for(i = 0; i < p_svet->n_object_num; i ++) {
		if(!(p_object = p_MergeVertices(&p_svet->object[i]))) {
			printf("    - Failure\n");
			return;
			getchar();
		}
		// vytvo objekt s izolovanmi vertexy (kad face m svoje vlastn vertexy)

		free((void*)p_svet->object[i].face);
		free((void*)p_svet->object[i].vertex);
		// uvoln stvajc vertexy a facy

		p_svet->object[i].face = p_object->face;
		p_svet->object[i].n_vertex_num = p_object->n_vertex_num;
		p_svet->object[i].vertex = p_object->vertex;
		// nastav nov vertexy a jejich poet a facy (poet fac se nezmn)

		free((void*)p_object);
		// uvoln pomocn objekt
	}*/
	// slou vertexy se stejnmi souadnicemi ...

	printf("\n- Writing lightmaps -\n");

	for(i = 0; i < n_lightmap_page_num; i ++) {
		sprintf(str, "_lightmap_pg_%s_%04d_.bmp", argv[1], i);
		//
		tmp_bmp.n_width = p_lightmap_page[i]->n_width;
		tmp_bmp.n_height = p_lightmap_page[i]->n_height;
		tmp_bmp.p_bytes = p_lightmap_page[i]->p_bytes;
		//
		if(!Save_TrueColor_BMP(str, &tmp_bmp)) {
			printf("    - Failure\n");
			return;
			getchar();
		}
	}

	printf("- Writing texcoords -\n");

	strcpy(str, argv[1]);
	if(strrchr(str, '.'))
		strcpy(strrchr(str, '.'), ".mesh");
	else
		strcat(str, ".mesh");
	// vyrob jmno vstupnho souboru ...

	if(!(fw = fopen(str, "wb"))) {
		printf("    - Failure\n");
		return;
		getchar();
	}
	fwrite(&p_svet->n_object_num, sizeof(int), 1, fw);
	fwrite(&n_lightmap_page_num, sizeof(int), 1, fw);
	fwrite(&p_svet->n_material_num, sizeof(int), 1, fw);
	// zape poet objekt a poet strnek lightmap ...

	for(i = 0; i < n_lightmap_page_num; i ++) {
		sprintf(str, "_lightmap_pg_%s_%04d_.bmp", argv[1], i);
		n = strlen(str);
		fwrite(&n, sizeof(int), 1, fw);
		fwrite(str, sizeof(char), n, fw);
	}
	// zape jmna strnek s lightmapama (bez ukonovac nuly)

	for(i = 0; i < p_svet->n_material_num; i ++) {
		n = strlen(p_svet->material[i].name);
		fwrite(&n, sizeof(int), 1, fw);
		fwrite(p_svet->material[i].name, sizeof(char), n, fw);
	}
	// zape jmna materil (bez ukonovac nuly)

	for(i = 0; i < p_svet->n_object_num; i ++) {
		fwrite(&p_svet->object[i].n_vertex_num, sizeof(int), 1, fw);
		fwrite(&p_svet->object[i].n_face_num, sizeof(int), 1, fw);
		// zape poty fac a vertex ...

		fwrite(p_svet->object[i].vertex, sizeof(Vertex), p_svet->object[i].n_vertex_num, fw);
		// zape vertexy ...

		for(j = 0; j < p_svet->object[i].n_face_num; j ++) {
			fwrite(&p_svet->object[i].face[j].n_lightmap, sizeof(int), 1, fw);
			fwrite(&p_svet->object[i].face[j].n_material, sizeof(int), 1, fw);
			//
			for(l = 0; l < 3; l ++) {
				n = p_svet->object[i].face[j].vertex[l] - p_svet->object[i].vertex;
				fwrite(&n, sizeof(int), 1, fw);
			}
			//
			fwrite(&p_svet->object[i].face[j].normal, sizeof(Plane), 1, fw);
		}
		// zape facy ...
	}
	// zape celou geometrii svta

	fclose(fw);

	printf("- done\n");
	getchar();
}